//
// Created by philipp on 12.04.18.
//

#include "../include/PlatoonProtocolLib/Protocol.h"
#include "../include/PlatoonProtocolLib/FollowerVehicle.h"
#include "../include/PlatoonProtocolLib/LeaderVehicle.h"
#include "NetworkingLib/Networking.h"
#include <chrono>

namespace platoonProtocol
{

networking::Networking net;
const std::string broadcastAddress = "192.168.2.255";
auto leader = LeaderVehicle::create(net, NetworkInfo{1, broadcastAddress});
auto follower1 = FollowerVehicle::create(net, NetworkInfo{2, broadcastAddress});
auto follower2 = FollowerVehicle::create(net, NetworkInfo{3, broadcastAddress});

using namespace std::chrono_literals;
const auto DEFAULT_WAIT_TOLERANCE = 30s;
const auto TIMEOUT_TOLERANCE = 10ms;

const PlatoonConfig TEST_CONFIG{1.0f, 10.0f};

networking::time::Duration abs(networking::time::Duration d)
{
    return d >= d.zero() ? d : -d;
}

bool waitUntil(networking::time::Duration maxWaitDuration, const std::function<bool()> & condition)
{
    auto startTime = networking::time::now();
    while (!condition() && networking::time::now() - startTime < maxWaitDuration);
    return condition() && networking::time::now() - startTime < maxWaitDuration;
}

// -------------------
// Test 1
// -------------------

bool test1Leader()
{
    leader->setPlatoonConfig(TEST_CONFIG);
    leader->createPlatoon();
    return waitUntil(DEFAULT_WAIT_TOLERANCE, [&]
    { return leader->isPlatoonRunning(); });
}

bool test1Follower1()
{
    follower1->createPlatoon();
    return waitUntil(DEFAULT_WAIT_TOLERANCE, [&]
    { return follower1->isPlatoonRunning(); });
}

bool test1Follower2()
{
    return true;
}

// -------------------
// Test 2
// -------------------

bool test2Leader()
{
    if (!test1Leader())
        return false;
    // Give some time to send the first broadcast message.
    std::this_thread::sleep_for(5s);
    return true;
}

bool test2Follower1()
{
    follower1->createPlatoon();
    return waitUntil(DEFAULT_WAIT_TOLERANCE, [&]
    {
        auto config = follower1->getPlatoonConfig();
        std::cout << "running: " << follower1->isPlatoonRunning() << "\n";
        std::cout << "ps: " << config.platoonSpeed << " ipd: " << config.innerPlatoonDistance << "\n";
        return follower1->isPlatoonRunning() &&
               follower1->getPlatoonConfig() == TEST_CONFIG;
    });
}

bool test2Follower2()
{
    return true;
}

// -------------------
// Test 3
// -------------------

bool test3Leader()
{
    if (!test1Leader())
        return false;

    // Wait a little bit so the first config can be send.
    std::this_thread::sleep_for(10ms);

    for (std::size_t i = 0; i < 10; i++)
    {
        leader->setPlatoonConfig(PlatoonConfig{TEST_CONFIG.platoonSpeed + i * 0.1f,
                                               TEST_CONFIG.innerPlatoonDistance + i});
        std::this_thread::sleep_for(LeaderVehicle::BROADCAST_INTERVAL);
    }

    // Give some time to send the last broadcast message.
    std::this_thread::sleep_for(50ms);
}

bool test3Follower1()
{
    std::atomic<bool> running{true};
    std::atomic<bool> success{false};
    auto expectedConfig = TEST_CONFIG;
    auto lastConfigTime = networking::time::TimePoint::min();
    std::size_t numConfigsReceived = 0;

    follower1->setOnPlatoonConfigUpdatedCallback(
        [&]
        {
            auto deltaTime = networking::time::now() - lastConfigTime;

            if (follower1->getPlatoonConfig() != expectedConfig ||
                (numConfigsReceived > 0 &&
                 abs(deltaTime - LeaderVehicle::BROADCAST_INTERVAL) > TIMEOUT_TOLERANCE))
            {
                success = false;
                running = false;
                return;
            }

            expectedConfig.platoonSpeed += 0.1f;
            expectedConfig.innerPlatoonDistance += 1.0f;
            numConfigsReceived++;
            if (numConfigsReceived >= 10)
            {
                success = true;
                running = false;
            }
        });

    follower1->createPlatoon();

    return waitUntil(DEFAULT_WAIT_TOLERANCE, [&]
    { return !running; }) && success;
}

bool test3Follower2()
{
    return true;
}

// -------------------
// Test 4
// -------------------

bool test4Leader()
{
    if (!test1Leader())
        return false;
    std::this_thread::sleep_for(1s);
    leader->stop();
    return true;
}

bool test4Follower1()
{
    if (!test1Follower1())
        return false;
    return waitUntil(1s + FollowerVehicle::BROADCAST_TIMEOUT + TIMEOUT_TOLERANCE, [&]
    { return !follower1->isPlatoonRunning(); });
}

bool test4Follower2()
{
    return true;
}

// -------------------
// Test 5
// -------------------

bool test5Leader()
{
    if (!test1Leader())
        return false;
    return waitUntil(1s + LeaderVehicle::HEARTBEAT_TIMEOUT + TIMEOUT_TOLERANCE, [&]
    { return !leader->isPlatoonRunning(); });
}

bool test5Follower1()
{
    if (!test1Follower1())
        return false;
    std::this_thread::sleep_for(1s);
    follower1->stop();
    return true;
}

bool test5Follower2()
{
    return true;
}

// -------------------
// Test 6
// -------------------

bool test6Leader()
{
    if (!test1Leader())
        return false;
    return waitUntil(1100ms, [&]
    { return !leader->isPlatoonRunning(); });
}

bool test6Follower1()
{
    if (!test1Follower1())
        return false;
    std::this_thread::sleep_for(1s);
    follower1->leavePlatoon();
    // Give some time to send leave platoon message.
    std::this_thread::sleep_for(50ms);
    return true;
}

bool test6Follower2()
{
    return true;
}

// -------------------
// Test 7
// -------------------

bool test7Leader()
{
    return true;
}

bool test7Follower1()
{
    follower1->createPlatoon();
    std::this_thread::sleep_for(5s);
    return !follower1->isPlatoonRunning();
}

bool test7Follower2()
{
    follower2->createPlatoon();
    std::this_thread::sleep_for(5s);
    return !follower2->isPlatoonRunning();
}

// -------------------
// Test 8
// -------------------

bool test8Leader()
{
    if (!test1Leader())
        return false;
    std::this_thread::sleep_for(100ms);
    leader->setPlatoonConfig(PlatoonConfig{TEST_CONFIG.platoonSpeed + 0.1f, TEST_CONFIG.innerPlatoonDistance + 1.0f});
    std::this_thread::sleep_for(100ms);
    leader->setPlatoonConfig(PlatoonConfig{TEST_CONFIG.platoonSpeed + 0.2f, TEST_CONFIG.innerPlatoonDistance + 2.0f});
    std::this_thread::sleep_for(100ms);
    return true;
}

bool test8Followers(FollowerVehicle::Ptr follower)
{
    std::atomic<bool> success{false};
    PlatoonConfig expectedConfig = TEST_CONFIG;
    std::size_t numConfigsReceived = 0;

    follower->setOnPlatoonConfigUpdatedCallback(
        [&]
        {
            if (follower->getPlatoonConfig() != expectedConfig)
            {
                success = false;
                return;
            }

            expectedConfig.platoonSpeed += 0.1f;
            expectedConfig.innerPlatoonDistance += 1.0f;
            numConfigsReceived++;
            if (numConfigsReceived >= 3)
                success = true;
        });

    follower->createPlatoon();

    return waitUntil(1s, [&]
    { return success.load(); } );
}

bool test8Follower1()
{
    return test8Followers(follower1);
}

bool test8Follower2()
{
    return test8Followers(follower2);
}

// -------------------
// Test 9
// -------------------

bool test9Leader()
{
    if (!test1Leader())
        return false;
    std::this_thread::sleep_for(1100ms);
    bool result = leader->isPlatoonRunning();
    // Wait some time to signal that the platoon is still running.
    std::this_thread::sleep_for(900ms);
    return result;
}

bool test9Follower1()
{
    if (!test1Follower1())
        return false;
    std::this_thread::sleep_for(1s);
    follower1->leavePlatoon();
    // Give some time to send the leave platoon message.
    std::this_thread::sleep_for(50ms);
    return true;
}

bool test9Follower2()
{
    follower2->createPlatoon();
    auto result = waitUntil(DEFAULT_WAIT_TOLERANCE, [&]
    { return follower2->isPlatoonRunning(); });
    if (!result)
        return false;
    std::this_thread::sleep_for(1100ms);
    result = follower2->isPlatoonRunning();
    // Wait some time to signal that the platoon is still running.
    std::this_thread::sleep_for(900ms);
    return result;
}

bool runTestScenario(int testNum, int testRole)
{
    using Test = bool (*)();
    Test tests[9][3] = {
        {test1Leader, test1Follower1, test1Follower2},
        {test2Leader, test2Follower1, test2Follower2},
        {test3Leader, test3Follower1, test3Follower2},
        {test4Leader, test4Follower1, test4Follower2},
        {test5Leader, test5Follower1, test5Follower2},
        {test6Leader, test6Follower1, test6Follower2},
        {test7Leader, test7Follower1, test7Follower2},
        {test8Leader, test8Follower1, test8Follower2},
        {test9Leader, test9Follower1, test9Follower2}
    };
    return tests[testNum - 1][testRole - 1]();
}

}

int main(int argc, char ** argv)
{
    if (argc < 3)
    {
        std::cerr << "Too few arguments!\n";
        return -1;
    }

    int testNum = atoi(argv[1]);
    int testRole = atoi(argv[2]);

    if (testNum < 1 || testNum > 9)
    {
        std::cerr << "First argument must be between 1 and 9!\n";
        return -1;
    }

    if (testRole < 1 || testRole > 3)
    {
        std::cerr << "Second argument must be between 1 and 3";
        return -1;
    }

    auto result = platoonProtocol::runTestScenario(testNum, testRole);
    if (result)
        std::cout << "1";
    else
        std::cout << "0";

    return 0;
}